/*
	bfsha1 - Brute forces a SHA1 hash
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <stdarg.h>
#include "common.h"

void reportError(const char *msg, ...)
{
	va_list args;
	char *format;
	size_t len = strlen(msg);

	format = new char[len + 9]; // 9 == sizeof("Error: \n")
	strcpy(format, "Error: ");
	strcpy(format + 7, msg);
	format[len + 7] = '\n';
	format[len + 8] = 0;
	va_start(args, msg);
	vfprintf(stderr, format, args);
	va_end(args);
	delete [] format;
}

void reportWarning(const char *msg, ...)
{
	va_list args;
	char *format;
	size_t len = strlen(msg);

	format = new char[len + 11]; // 11 == sizeof("Warning: \n")
	strcpy(format, "Warning: ");
	strcpy(format + 9, msg);
	format[len + 9] = '\n';
	format[len + 10] = 0;
	va_start(args, msg);
	vfprintf(stderr, format, args);
	va_end(args);
	delete [] format;
}

float reciprocalRoundDown(uint64_t num)
{
	// keySpace = (36 ^ 8 - 36) / 35
	// 1.0f / keySpace = 14,303,679.752109742480909981844964 / 2 ^ 60 = 14,303,680 / 2 ^ 60
	// floor(14,303,680 / 2 ^ 60 * (keySpace - 1)) = floor(1.0000000173181169960123781947914) = 1
	// Very very bad since:
	// (keySpace - 1) - keySpace * 1 = -1 = 0xffffffffffffffff = 18,446,744,073,709,551,615 >= keySpace

	// round_down(1.0f / keySpace) =
	// round_down(14,303,679.752109742480909981844964 / 2 ^ 60) =
	// 14,303,679 / 2 ^ 60

	// keySpaceReciprocal = round_down(1.0f / m_keySpace)
	// Note that setting the rounding mode to round down may not work correctly
	// if the key space is > 2 ^ 54 well unless you can be sure that 

	// This may not work sometimes unless you change the [under|over]41bits functions to [under|over]42bits
	// I think the limit is under/over 2 ^ 40 but made it 2 ^ 41 just incase so then the limit of 2 ^ 41 would work
	// float keySpaceReciprocal = 1.0 / keySpace;
	// uint32_t tmp = (*((uint32_t*) &keySpaceReciprocal) & 0xff800000) - (23 << 23); // :)
	// keySpaceReciprocal -= *((float*) &tmp);

	// I think this should work:
	// #ifdef _WIN32
	//     #include <fenv.h>
	//     #pragma STDC FENV_ACCESS ON
	//     int tmp = fegetround();
	//     fesetround(FE_DOWNWARD);
	//     float keySpaceReciprocal = 1.0 / keySpace;
	//     fesetround(tmp);
	// #else
	//     unsigned int tmp = _control_fp(0, 0);
	//     _control_fp(_RC_DOWN, _MCW_RC);
	//     float keySpaceReciprocal = 1.0 / keySpace;
	//     _control_fp(tmp, _MCW_RC);
	// #endif

	// I totally guarantee this to find the keySpaceReciprocal I'm looking for
	uint64_t q, r;
	float reciprocal;

	if (num == 0)
	{
		//reciprocal = 1.0f / 0.0f;
		reciprocal = 0.0f;
	}
	else if (num == 1)
	{
		reciprocal = 1.0f;
	}
	else
	{
		q = 0xffffffffffffffff / num;
		r = 0xffffffffffffffff % num + 1;
		if (r == num)
		{
			q++;
			r = 0;
		}
		uint32_t power = countLeadingZeros(q) + 1;
		uint32_t highestBit = 65 - power;
		if (highestBit > 24)
		{
			q >>= highestBit - 24;
		}
		else
		{
			while (q <= 8388608)
			{
				q <<= 1;
				r <<= 1;
				if (r >= num)
				{
					q++;
					r -= num;
				}
			}
		}
		*((uint32_t*) &reciprocal) = (((uint32_t) q) & 0x007fffff) + ((127 - power) << 23);
	}
	return reciprocal;
}
