/*
	bfsha1 - Brute forces a SHA1 hash
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "common.h"
#include "gpucracker.h"
#include <stdlib.h>
#include <errno.h>

#define USE_STR_TO_UL
#define USE_STR_TO_ULL

int getHash(const char *hash, uint32_t *out);
void usage(const char *runStr);
int validateInt(const char *arg, uint32_t &out, bool required, const char *name, uint32_t rangeLo = 1, uint32_t rangeHi = 0, uint32_t defaultVal = 0, char endChar = 0, char **pos = NULL);

int main(int argc, char *argv[])
{
	uint64_t maxIndex = UINT64_C(6634204312890625); // 95 ^ 8
	uint64_t count, index;
	TIMER_TYPE s, e;
	GpuCracker *cracker;
	int benchmark = 0;

	// *** Get args ***
	const char *hashIn;
	uint32_t gpuDeviceNum = 0, hash[5] = {0};

	if (argc < 2 || argc > 5)
	{
		usage(argv[0]);
		return 1;
	}
	if (STR_CMP_ICASE(argv[1], "--benchmark") == 0)
	{
		benchmark = 1;
	}
	else
	{
		hashIn = argv[1];
		if (getHash(hashIn, hash))
		{
			reportError("Invalid SHA1 hash \"%s\".", hashIn);
			return 1;
		}
	}
	if (argc == 3 || argc == 5)
	{
		if (validateInt(argv[2], gpuDeviceNum, true, "gpu-device-num", 0, 65536))
		{
			return 1;
		}
	}
	if (argc == 4 || argc == 5)
	{
		uint32_t blocks, threads;

		if (validateInt(argv[argc - 2], blocks, true, "gpu-blocks", 1, 65536))
		{
			return 1;
		}
		if (validateInt(argv[argc - 1], threads, true, "gpu-threads", 1, 65536))
		{
			return 1;
		}
		setDefaultBlocksThreads(gpuDeviceNum, (int) blocks, (int) threads);
	}
	else
	{
		setDefaultThreads(gpuDeviceNum, 512);
	}

	// *** Do stuff ***
	int blocks, threads;

	getDefaultBlocksThreads(gpuDeviceNum, blocks, threads);
	printf("Blocks * threads: %d * %d\n", blocks, threads);

	cracker = new GpuCracker(" !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~", 95, 8, hash, "sha1", gpuDeviceNum);
	count = cracker->getPwsPerCall();

	if (benchmark)
	{
		maxIndex = 60 * count;
	}

	TIMER_FUNC(s);
	for (index = 0; index < maxIndex; index += count)
	{
		if (cracker->crack(index))
		{
			index += count;
			break;
		}
	}
	TIMER_FUNC(e);
	cracker->printPw();

	// *** Report ***
	if (TIMER_DIFF(s, e) < 100.0)
	{
		printf("%" PRIu64 " passwords in %0.3f seconds @ %0.2f M/sec\n", index, TIMER_DIFF(s, e), index / TIMER_DIFF(s, e) / 1000000.0);
	}
	else if (TIMER_DIFF(s, e) < 1000.0)
	{
		printf("%" PRIu64 " passwords in %0.2f seconds @ %0.2f M/sec\n", index, TIMER_DIFF(s, e), index / TIMER_DIFF(s, e) / 1000000.0);
	}
	else if (TIMER_DIFF(s, e) < 10000.0)
	{
		printf("%" PRIu64 " passwords in %0.1f seconds @ %0.2f M/sec\n", index, TIMER_DIFF(s, e), index / TIMER_DIFF(s, e) / 1000000.0);
	}
	else
	{
		printf("%" PRIu64 " passwords in %0.0f seconds @ %0.2f M/sec\n", index, TIMER_DIFF(s, e), index / TIMER_DIFF(s, e) / 1000000.0);
	}
//	getchar();
	return 0;
}

int getHash(const char *hash, uint32_t *out)
{
	uint32_t tmp = 0;

	if (hash[8 * 5] != 0)
	{
		return 1;
	}
	for (uint32_t i = 0; i < 5; i++)
	{
		for (uint32_t j = 8 * i; j < 8 * i + 8; j++)
		{
			char offset = 'A' - 10;

			if (hash[j] >= '0' && hash[j] <= '9')
			{
				offset = '0';
			}
			else if (hash[j] >= 'a' && hash[j] <= 'f')
			{
				offset = 'a' - 10;
			}
			else if (hash[j] < 'A' || hash[j] > 'F')
			{
				return 1;
			}
			tmp <<= 4;
			tmp |= hash[j] - offset;
		}
		out[i] = tmp;
	}
	return 0;
}

int validateInt(const char *arg, uint32_t &out, bool required, const char *name, uint32_t rangeLo, uint32_t rangeHi, uint32_t defaultVal, char endChar, char **pos)
{
#ifdef USE_STR_TO_UL
	char *endPtr;
#endif
	uint32_t val = 0;
	int ret = 0;

	if (arg == NULL)
	{
		out = defaultVal;
		if (required)
		{
			reportError("Missing argument %s.", name);
			ret = 1;
		}
	}
	else
	{
#ifdef USE_STR_TO_UL
		errno = 0;
		out = val = (uint32_t) strtoul(arg, &endPtr, 0);
		if ((*endPtr != endChar && *endPtr != 0) || errno != 0)
		{
			ret = 1;
		}
		if (pos != NULL)
		{
			*pos = endPtr;
		}
#else
		if (*arg == '0')
		{
			if (arg[1] == 'x')
			{
				for (arg += 2; *arg != endChar || *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '9')
					{
						val = 16 * val + (uint32_t) (*arg - '0');
					}
					else if (*arg >= 'a' && *arg <= 'f')
					{
						val = 16 * val + (uint32_t) (*arg - 'a') + 10;
					}
					else if (*arg >= 'A' && *arg <= 'F')
					{
						val = 16 * val + (uint32_t) (*arg - 'A') + 10;
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
			else
			{
				for (arg++; *arg != endChar || *arg != 0; arg++)
				{
					if (*arg >= '0' && *arg <= '7')
					{
						val = 8 * val + (uint32_t) (*arg - '0');
					}
					else
					{
						ret = 1;
						break;
					}
				}
			}
		}
		else
		{
			for (; *arg != endChar || *arg != 0; arg++)
			{
				if (*arg >= '0' && *arg <= '9')
				{
					val = 10 * val + (uint32_t) (*arg - '0');
				}
				else
				{
					ret = 1;
					break;
				}
			}
		}
		if (pos != NULL)
		{
			*pos = arg;
		}
#endif
		if (rangeLo > rangeHi)
		{
			if (ret)
			{
				reportError("Invalid argument %s must be an integer.", name);
			}
		}
		else if (ret || val < rangeLo || val > rangeHi)
		{
			reportError("Invalid argument %s must be an integer between %u and %u.", name, rangeLo, rangeHi);
			ret = 1;
		}
	}
	return ret;
}

void usage(const char *runStr)
{
	printf("\n"
		"                       **** Brute force SHA1 v0.1.2a ****\n\n"
		"Brute forces a SHA1 hash with a key space of [ -~]{8}.\n\n"
//		"%s hash [threads [gpu-device-num [...]]]\n\n"
		"%s {--benchmark|hash} [gpu-device-num] [gpu-blocks gpu-threads]\n\n"
		"hash:\n"
		"  A SHA1 hash.\n\n"
/*		"threads:\n"
		"  Number of CPU threads.\n\n"*/
		"gpu-device-num:\n"
		"  CUDA GPU device number (starts at 0).\n\n"
		"Copyright (c) 2012, Steve Thomas <steve AT tobtu DOT com>\n"
		"All rights reserved.\n\n"
		"  This is free software: you can redistribute it and/or modify it under the\n"
		"  terms of the GNU General Public License as published by the Free Software\n"
		"  Foundation, either version 2 of the License, or (at your option) any later\n"
		"  version. There is NO WARRANTY, to the extent permitted by law.\n", runStr);
}
