/*
	bfsha1 - Brute forces a SHA1 hash
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <cuda.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#ifndef uint8_t
	#define uint8_t unsigned char
#endif
#ifndef uint16_t
	#define uint16_t unsigned short
#endif
#ifndef uint32_t
	#define uint32_t unsigned int
#endif
#ifndef uint64_t
	#define uint64_t unsigned long long int
#endif

typedef void (*kernel_cracker)(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex);

// 64 KiB of constant memory
// 16 Bytes
__constant__ uint32_t c_charSetLength;
__constant__ float    c_charSetReciprocal;
__constant__ float    c_charSetLen;
__constant__ float    c_ncharSetLen;

// 28 Bytes
__constant__ uint32_t c_ha;
__constant__ uint32_t c_hb;
__constant__ uint32_t c_hc;
__constant__ uint32_t c_hd;
__constant__ uint32_t c_he;
__constant__ uint32_t c_dXORe;
__constant__ uint32_t c_heOld;

// 41 * 4 * 192 = 61.5 KiB
__constant__ uint32_t c_wRounds[2*192*41];
#define c_char(i)   c_wRounds[0*192*41 + 192* 0 + i]
#define c_r17(i)    c_wRounds[0*192*41 + 192* 1 + i]
#define c_r20(i)    c_wRounds[0*192*41 + 192* 2 + i]
#define c_r23(i)    c_wRounds[0*192*41 + 192* 3 + i]
#define c_r26(i)    c_wRounds[0*192*41 + 192* 4 + i]
#define c_r29(i)    c_wRounds[0*192*41 + 192* 5 + i]
#define c_r31(i)    c_wRounds[0*192*41 + 192* 6 + i]
#define c_r32(i)    c_wRounds[0*192*41 + 192* 7 + i]
#define c_r33(i)    c_wRounds[0*192*41 + 192* 8 + i]
#define c_r35(i)    c_wRounds[0*192*41 + 192* 9 + i]
#define c_r37(i)    c_wRounds[0*192*41 + 192*10 + i]
#define c_r38(i)    c_wRounds[0*192*41 + 192*11 + i]
#define c_r41(i)    c_wRounds[0*192*41 + 192*12 + i]
#define c_r43(i)    c_wRounds[0*192*41 + 192*13 + i]
#define c_r44(i)    c_wRounds[0*192*41 + 192*14 + i]
#define c_r45(i)    c_wRounds[0*192*41 + 192*15 + i]
#define c_r47(i)    c_wRounds[0*192*41 + 192*16 + i]
#define c_r48(i)    c_wRounds[0*192*41 + 192*17 + i]
#define c_r49(i)    c_wRounds[0*192*41 + 192*18 + i]
#define c_r50(i)    c_wRounds[0*192*41 + 192*19 + i]
#define c_r53(i)    c_wRounds[0*192*41 + 192*20 + i]
#define c_r55(i)    c_wRounds[0*192*41 + 192*21 + i]
#define c_r56(i)    c_wRounds[0*192*41 + 192*22 + i]
#define c_r57(i)    c_wRounds[0*192*41 + 192*23 + i]
#define c_r59(i)    c_wRounds[0*192*41 + 192*24 + i]
#define c_r60(i)    c_wRounds[0*192*41 + 192*25 + i]
#define c_r61(i)    c_wRounds[0*192*41 + 192*26 + i]
#define c_r62(i)    c_wRounds[0*192*41 + 192*27 + i]
#define c_r63(i)    c_wRounds[0*192*41 + 192*28 + i]
#define c_r65(i)    c_wRounds[0*192*41 + 192*29 + i]
#define c_r67(i)    c_wRounds[0*192*41 + 192*30 + i]
#define c_r68(i)    c_wRounds[0*192*41 + 192*31 + i]
#define c_r69(i)    c_wRounds[0*192*41 + 192*32 + i]
#define c_r71(i)    c_wRounds[0*192*41 + 192*33 + i]
#define c_r72(i)    c_wRounds[0*192*41 + 192*34 + i]
#define c_r73(i)    c_wRounds[0*192*41 + 192*35 + i]
#define c_r74(i)    c_wRounds[0*192*41 + 192*36 + i]
#define c_r75(i)    c_wRounds[0*192*41 + 192*37 + i]
#define c_r76(i)    c_wRounds[0*192*41 + 192*38 + i]
#define c_r77(i)    c_wRounds[0*192*41 + 192*39 + i]
#define c_r79(i)    c_wRounds[0*192*41 + 192*40 + i]

#define c_char_2(i) c_wRounds[1*192*41 + 192* 0 + i]
#define c_r17_2(i)  c_wRounds[1*192*41 + 192* 1 + i]
#define c_r20_2(i)  c_wRounds[1*192*41 + 192* 2 + i]
#define c_r23_2(i)  c_wRounds[1*192*41 + 192* 3 + i]
#define c_r26_2(i)  c_wRounds[1*192*41 + 192* 4 + i]
#define c_r29_2(i)  c_wRounds[1*192*41 + 192* 5 + i]
#define c_r31_2(i)  c_wRounds[1*192*41 + 192* 6 + i]
#define c_r32_2(i)  c_wRounds[1*192*41 + 192* 7 + i]
#define c_r33_2(i)  c_wRounds[1*192*41 + 192* 8 + i]
#define c_r35_2(i)  c_wRounds[1*192*41 + 192* 9 + i]
#define c_r37_2(i)  c_wRounds[1*192*41 + 192*10 + i]
#define c_r38_2(i)  c_wRounds[1*192*41 + 192*11 + i]
#define c_r41_2(i)  c_wRounds[1*192*41 + 192*12 + i]
#define c_r43_2(i)  c_wRounds[1*192*41 + 192*13 + i]
#define c_r44_2(i)  c_wRounds[1*192*41 + 192*14 + i]
#define c_r45_2(i)  c_wRounds[1*192*41 + 192*15 + i]
#define c_r47_2(i)  c_wRounds[1*192*41 + 192*16 + i]
#define c_r48_2(i)  c_wRounds[1*192*41 + 192*17 + i]
#define c_r49_2(i)  c_wRounds[1*192*41 + 192*18 + i]
#define c_r50_2(i)  c_wRounds[1*192*41 + 192*19 + i]
#define c_r53_2(i)  c_wRounds[1*192*41 + 192*20 + i]
#define c_r55_2(i)  c_wRounds[1*192*41 + 192*21 + i]
#define c_r56_2(i)  c_wRounds[1*192*41 + 192*22 + i]
#define c_r57_2(i)  c_wRounds[1*192*41 + 192*23 + i]
#define c_r59_2(i)  c_wRounds[1*192*41 + 192*24 + i]
#define c_r60_2(i)  c_wRounds[1*192*41 + 192*25 + i]
#define c_r61_2(i)  c_wRounds[1*192*41 + 192*26 + i]
#define c_r62_2(i)  c_wRounds[1*192*41 + 192*27 + i]
#define c_r63_2(i)  c_wRounds[1*192*41 + 192*28 + i]
#define c_r65_2(i)  c_wRounds[1*192*41 + 192*29 + i]
#define c_r67_2(i)  c_wRounds[1*192*41 + 192*30 + i]
#define c_r68_2(i)  c_wRounds[1*192*41 + 192*31 + i]
#define c_r69_2(i)  c_wRounds[1*192*41 + 192*32 + i]
#define c_r71_2(i)  c_wRounds[1*192*41 + 192*33 + i]
#define c_r72_2(i)  c_wRounds[1*192*41 + 192*34 + i]
#define c_r73_2(i)  c_wRounds[1*192*41 + 192*35 + i]
#define c_r74_2(i)  c_wRounds[1*192*41 + 192*36 + i]
#define c_r75_2(i)  c_wRounds[1*192*41 + 192*37 + i]
#define c_r76_2(i)  c_wRounds[1*192*41 + 192*38 + i]
#define c_r77_2(i)  c_wRounds[1*192*41 + 192*39 + i]
#define c_r79_2(i)  c_wRounds[1*192*41 + 192*40 + i]

/*
__constant__ uint32_t c_char[192];
__constant__ uint32_t c_char_2[192];
// c_r20 = "c_r25"
// c_r26 = "c_r36" = "c_r39"
// c_r37 = "c_r52"
// c_r38 = "c_r51" = "c_r58" = "c_r64"
__constant__ uint32_t c_r17[192],   c_r20[192],   c_r23[192],   c_r26[192],   c_r29[192],   c_r31[192],   c_r32[192],   c_r33[192],   c_r35[192],   c_r37[192];
__constant__ uint32_t c_r38[192],   c_r41[192],   c_r43[192],   c_r44[192],   c_r45[192],   c_r47[192],   c_r48[192],   c_r49[192],   c_r50[192],   c_r53[192];
__constant__ uint32_t c_r55[192],   c_r56[192],   c_r57[192],   c_r59[192],   c_r60[192],   c_r61[192],   c_r62[192],   c_r63[192],   c_r65[192],   c_r67[192];
__constant__ uint32_t c_r68[192],   c_r69[192],   c_r71[192],   c_r72[192],   c_r73[192],   c_r74[192],   c_r75[192],   c_r76[192],   c_r77[192],   c_r79[192];
__constant__ uint32_t c_r17_2[192], c_r20_2[192], c_r23_2[192], c_r26_2[192], c_r29_2[192], c_r31_2[192], c_r32_2[192], c_r33_2[192], c_r35_2[192], c_r37_2[192];
__constant__ uint32_t c_r38_2[192], c_r41_2[192], c_r43_2[192], c_r44_2[192], c_r45_2[192], c_r47_2[192], c_r48_2[192], c_r49_2[192], c_r50_2[192], c_r53_2[192];
__constant__ uint32_t c_r55_2[192], c_r56_2[192], c_r57_2[192], c_r59_2[192], c_r60_2[192], c_r61_2[192], c_r62_2[192], c_r63_2[192], c_r65_2[192], c_r67_2[192];
__constant__ uint32_t c_r68_2[192], c_r69_2[192], c_r71_2[192], c_r72_2[192], c_r73_2[192], c_r74_2[192], c_r75_2[192], c_r76_2[192], c_r77_2[192], c_r79_2[192];
*/
//#if defined(__CUDA_ARCH__) && __CUDA_ARCH__ >= 200
	#define USE___fmul_rd
//#else
	// none
//#endif

extern __shared__ uint32_t s_shared[];

template <int SIZE,int CHEAT>
__global__ static void sbSha1_cracker_any(uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
#define ROL(x,s) (((x) << (s)) + ((x) >> (32 - (s))))
	// 0 init
	// 1 index to pw
	// 2 generate w[16-79]
	// 3 reverse hash
	// 4 precalc first steps
	// loop
	//  5 start hash
	//  6 hash
	//  7 finish hash
	//  8 check

	if (SIZE <= 5 || pwLen <= 5)
	{
		return;
	}

	// **** 0 init ****
	uint64_t *s_shortCut           = (uint64_t*)  s_shared;
	uint32_t *s_shortCutShift      = (uint32_t*) (s_shortCut           + 16);
	float    *s_shortCutReciprocal = (float*)    (s_shortCutShift      + 16);
	float    *s_shortCutf          = (float*)    (s_shortCutReciprocal + 16);
	uint8_t  *s_charSet            = (uint8_t*)  (s_shortCutf          +  3);
	const uint32_t THREAD_ID       = blockIdx.x * blockDim.x + threadIdx.x;

	for (uint32_t i = threadIdx.x; i < 128 + 3; i += blockDim.x)
	{
		s_shared[i] = d_shared[i];
	}
	__syncthreads();


	// **** 1 index to pw ****
	uint32_t pw0, pw1, pw2, pw3;
	float r, q;
	float index1, index2, index3, index4, index5;

	index += THREAD_ID;
	if (index >= maxIndex)
	{
		return;
	}

	// Split index (64 bit int) into 3 character chunks as floats: index1, index2, index3, index4, and index5
#ifdef USE___fmul_rd
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64) \
		indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
		index64 -= shortCut64 * ((uint32_t) indexOut); \
		if (index64 >= shortCut64) \
		{ \
			index64 -= shortCut64; \
			indexOut++; \
		}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
		indexOut = floorf(__fmul_rd((float) ((uint32_t) (index64 >> shortCutShift)), shortCutReciprocal)); \
		indexOut2 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
		if (indexOut2 >= shortCutf) \
		{ \
			indexOut2 -= shortCutf; \
			indexOut++; \
		}
#else
	uint32_t index32;
	#define SHORT_CUT(index64, indexOut, shortCutShift, shortCutReciprocal, shortCut64, shortCutf) \
		indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
		index64 -= shortCut64 * ((uint32_t) indexOut); \
		if (index64 >= 0x8000000000000000) \
		{ \
			index64 += shortCut64; \
			indexOut--; \
		} \
		else if (index64 >= shortCut64) \
		{ \
			index64 -= shortCut64; \
			indexOut++; \
		}
	#define SHORT_CUT_LAST(index64, indexOut, indexOut2, shortCutShift, shortCutReciprocal, shortCut32, shortCutf) \
		indexOut = floorf((float) ((uint32_t) (index64 >> shortCutShift)) * shortCutReciprocal); \
		index32 = (uint32_t) (index64 - (uint64_t) ((uint32_t) shortCut32) * ((uint32_t) indexOut)); \
		if (index32 >= 0x80000000) \
		{ \
			index32 += shortCut32; \
			indexOut--; \
		} \
		indexOut2 = index32; \
		if (indexOut2 >= shortCutf) \
		{ \
			indexOut2 -= shortCutf; \
			indexOut++; \
		}
#endif
	if (SIZE > 12 && pwLen > 12) // pw3 = (13, 14, 15, ...)
	{
		pw3 = 0x80000000;
		SHORT_CUT     (index, index1,         s_shortCutShift[pwLen- 3-1], s_shortCutReciprocal[pwLen- 3-1], s_shortCut[pwLen- 3-1]);
		SHORT_CUT     (index, index2,         s_shortCutShift[pwLen- 6-1], s_shortCutReciprocal[pwLen- 6-1], s_shortCut[pwLen- 6-1]);
		SHORT_CUT     (index, index3,         s_shortCutShift[pwLen- 9-1], s_shortCutReciprocal[pwLen- 9-1], s_shortCut[pwLen- 9-1]);
		SHORT_CUT_LAST(index, index4, index5, s_shortCutShift[pwLen-12-1], s_shortCutReciprocal[pwLen-12-1], s_shortCut[pwLen-12-1], s_shortCutf[pwLen-12-1]);
	}
	else if (SIZE > 9 && pwLen > 9) // pw2 = (..., 10, 11, 12)
	{
		if (SIZE >= 12 && pwLen == 12)
		{
			pw3 = 0x80000000;
		}
		else
		{
			if (SIZE >= 12)
			{
				pw3 = 0;
			}
			pw2 = 0x80000000;
		}
		SHORT_CUT     (index, index1,         s_shortCutShift[pwLen-3-1], s_shortCutReciprocal[pwLen-3-1], s_shortCut[pwLen-3-1]);
		SHORT_CUT     (index, index2,         s_shortCutShift[pwLen-6-1], s_shortCutReciprocal[pwLen-6-1], s_shortCut[pwLen-6-1]);
		SHORT_CUT_LAST(index, index3, index4, s_shortCutShift[pwLen-9-1], s_shortCutReciprocal[pwLen-9-1], s_shortCut[pwLen-9-1], s_shortCutf[pwLen-9-1]);
	}
	else if (SIZE > 6 && pwLen > 6) // pw1 = (..., 7, 8), pw2 = (9, ...)
	{
		if (SIZE >= 12) pw3 = 0;
		if (SIZE > 7 && pwLen > 7)
		{
			pw2 = 0x80000000;
		}
		else
		{
			if (SIZE >= 8)
			{
				pw2 = 0;
			}
			pw1 = 0x80000000;
		}
		SHORT_CUT     (index, index1,         s_shortCutShift[pwLen-3-1], s_shortCutReciprocal[pwLen-3-1], s_shortCut[pwLen-3-1]);
		SHORT_CUT_LAST(index, index2, index3, s_shortCutShift[pwLen-6-1], s_shortCutReciprocal[pwLen-6-1], s_shortCut[pwLen-6-1], s_shortCutf[pwLen-6-1]);
	}
	else if (SIZE > 3 && pwLen > 3) // pw0 = (..., 4), pw1 = (5, 6, ...)
	{
		if (SIZE >= 12) pw3 = 0;
		if (SIZE >=  8) pw2 = 0;
		pw1 = 0x80000000;
		SHORT_CUT_LAST(index, index1, index2, s_shortCutShift[pwLen-3-1], s_shortCutReciprocal[pwLen-3-1], s_shortCut[pwLen-3-1], s_shortCutf[pwLen-3-1]);
	}
	else // pw0 = (1, 2, 3, ...)
	{
		if (SIZE >= 12) pw3 = 0;
		if (SIZE >=  8) pw2 = 0;
		if (SIZE >=  4) pw1 = 0;
		pw0 = 0x80000000;
		index1 = (uint32_t) index;
	}
#undef SHORT_CUT_NORM
#undef SHORT_CUT

	// Convert index1, index2, index3, index4, and index5 into a password
#define GET_CHAR(size, index, pw, charSet, charSetLen, ncharSetLen, charSetReciprocal) \
	if (SIZE >= size && pwLen >= size) \
	{ \
		q = floorf(index * charSetReciprocal); \
		r = ncharSetLen * q + index; \
		if (r >= charSetLen) \
		{ \
			r -= charSetLen; \
			q++; \
		} \
		index = q; \
		pw = (pw >> 8) + (((uint32_t) charSet[(uint32_t) r]) << 24); \
	}
#define GET_CHAR_LAST(size, index, pw, charSet) \
	if (SIZE >= size && pwLen >= size) \
	{ \
		pw = (pw >> 8) + (((uint32_t) charSet[(uint32_t) index]) << 24); \
	}
#define GET_CHAR_NORM(size, index, pw)       GET_CHAR(size, index, pw, s_charSet, c_charSetLen, c_ncharSetLen, c_charSetReciprocal)
#define GET_CHAR_LAST_NORM(size, index, pw)  GET_CHAR_LAST(size,index,pw,s_charSet)
	pwLen -= 2;
	GET_CHAR_LAST_NORM(13,index5,pw3);
	GET_CHAR_NORM     (12,index5,pw3);
	GET_CHAR_NORM     (11,index5,pw3);
	GET_CHAR_LAST_NORM(10,index4,pw2);
	GET_CHAR_NORM     ( 9,index4,pw2);
	GET_CHAR_NORM     ( 8,index4,pw2);
	GET_CHAR_LAST_NORM( 7,index3,pw2);
	GET_CHAR_NORM     ( 6,index3,pw1);
	GET_CHAR_NORM     ( 5,index3,pw1);
	// Skip two characters
	pw1 >>= 16;
	GET_CHAR_LAST_NORM( 4,index2,pw0);
	GET_CHAR_NORM     ( 3,index2,pw0);
	GET_CHAR_NORM     ( 2,index2,pw0);
	GET_CHAR_LAST_NORM( 1,index1,pw0);
	pwLen = 8 * pwLen + 16;
#undef GET_CHAR
#undef GET_CHAR_LAST
#undef GET_CHAR_NORM
#undef GET_CHAR_LAST_NORM


	// **** 2 generate w[16-79] ****
	uint32_t w[64];

	if (SIZE >=  8)
	{
		w[ 0] = ROL(                pw2   ^ pw0  , 1);
	}
	else
	{
		w[ 0] = ROL(                        pw0  , 1);
	}
	if (SIZE >= 12)
	{
		w[ 1] = ROL(                pw3   ^ pw1 ^ c_char_2(char2Start), 1);
	}
	else
	{
		w[ 1] = ROL(                        pw1 ^ c_char_2(char2Start), 1);
	}
	if (SIZE >=  8)
	{
		w[ 2] = ROL(pwLen ^                 pw2  , 1);
	}
	else
	{
		w[ 2] = ROL(pwLen                        , 1);
	}
	if (SIZE >= 12)
	{
		w[ 3] = ROL(w[ 0] ^                 pw3  , 1);
	}
	else
	{
		w[ 3] = ROL(w[ 0]                        , 1);
	}
	w[ 4] = ROL(w[ 1]                        , 1);
	w[ 5] = ROL(w[ 2]                        , 1);
	w[ 6] = ROL(w[ 3]                        , 1);
	w[ 7] = ROL(w[ 4] ^ pwLen                , 1);
	w[ 8] = ROL(w[ 5] ^ w[ 0]                , 1);
	w[ 9] = ROL(w[ 6] ^ w[ 1]                , 1);
	w[10] = ROL(w[ 7] ^ w[ 2]                , 1);
	w[11] = ROL(w[ 8] ^ w[ 3]                , 1);
	w[12] = ROL(w[ 9] ^ w[ 4]                , 1);
	w[13] = ROL(w[10] ^ w[ 5] ^ pwLen        , 1);
	w[14] = ROL(w[11] ^ w[ 6] ^ w[ 0]        , 1);
	w[15] = ROL(w[12] ^ w[ 7] ^ w[ 1] ^ pwLen, 1);

	#pragma unroll
	for (uint32_t i = 16; i < 64; i++)
	{
		w[i] = ROL(w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16], 1);
	}


	// **** 3 rev hash ****
	uint32_t hb;

	// ha -= 0x67452301;
	// hb -= 0xefcdab89;
	// hc -= 0x98badcfe;
	// hd -= 0x10325476;
	// he -= 0xc3d2e1f0;

	// SHA1_II(b,c,d,e,a,    x,79);
	// hc = ROL(hc, 2);
	// ha -= (hc ^ (hd ^ he)) + 0xca62c1d6 + ROL(hb, 5);
	// // ha -= w[79-16] ^ c_r79(i);

	// SHA1_II(c,d,e,a,b,    x,78);
	// hd = ROL(hd, 2);
	// hb -= 0xca62c1d6 + ROL(hc, 5);
	// c_dXORe = hd ^ he;
	hb = c_hb - w[78-16];
	// // hb -= (c_dXORe ^ ha) + w[78-16];

	// SHA1_II(d,e,a,b,c,    x,77);
	// heOld = he = ROL(he, 2);
	// hc -= 0xca62c1d6 + ROL(hd, 5);
	// // hc -= (he ^ (ha ^ hb)) + (w[77-16] ^ c_r77(i));

	// SHA1_II(e,a,b,c,d,    x,76);
	// // ha = ROL(ha, 2);
	// hd -= 0xca62c1d6 + ROL(heOld, 5);
	// // hd -= (ha ^ (hb ^ hc)) + (w[76-16] ^ c_r76(i));

	// SHA1_II(a,b,c,d,e,    x,75);
	// // hb = ROL(hb, 2);
	// he -= 0xca62c1d6;
	// // he -= (hb ^ (hc ^ hd)) + (w[75-16] ^ c_r75(i)) + ROL(ha, 5);

	// ...

	// SHA1_II(d,e,a,b,c,    x,72);
	// // he = ROL(he, 2);


	// **** 4 precalc first steps ****
	uint32_t ta, tb, tc, td, te, eXORa;

	te = 0x9fb498b3 + pw0;
	td = 0x66b0cd0d + ROL(te, 5) + pw1; // pw1 = 0x0000????
	if (SIZE >=  8)
	{
		tc = ((0x22222222 & te) ^ 0x7bf36ae2) + 0xf33d5697 + pw2;
	}
	else
	{
		tc = ((0x22222222 & te) ^ 0x7bf36ae2) + 0xf33d5697;
	}
	te = ROL(te, 30);
	if (SIZE >= 12)
	{
		tb = 0xd675e47b + pw3;
	}
	else
	{
		tb = 0xd675e47b;
	}
	eXORa = te ^ 0x59d148c0;
	ta = 0xb453c259; // + pw4


	// **** loop ****
	uint32_t j = char2Start;

	while (1)
	{
		td += c_char_2(j); // c_char_2(j) = 0x00??0000
		for (uint32_t i = 0; i < c_charSetLength; i++)
		{
			// **** 5 start hash ****
			uint32_t a = ta, b = tb, c = tc, d = td, e = te;

			d += c_char(i); // w[1] (c_char(i) = 0x??000000)
			c += ROL(d, 5);
			b += ((eXORa & d) ^ 0x59d148c0) + ROL(c, 5);
			d = ROL(d, 30);

			a += (((d ^ e) & c) ^ e) + ROL(b, 5);
			c = ROL(c, 30);


			// **** 6 hash ****
			// c_r20 = c_r25
			// c_r26 = c_r36 = c_r39
			// c_r37 = c_r52
			// c_r38 = c_r51 = c_r58 = c_r64
	#define SHA1_FF(_a,_b,_c,_d,_e,x,i)      _e += (((_c ^ _d) & _b) ^ _d)        + (x) + 0x5a827999 + ROL(_a, 5); _b = ROL(_b, 30);
	#define SHA1_GG(_a,_b,_c,_d,_e,x,i)      _e += (_b ^ (_c ^ _d))               + (x) + 0x6ed9eba1 + ROL(_a, 5); _b = ROL(_b, 30);
	#define SHA1_HH(_a,_b,_c,_d,_e,x,i)      _e += (((_b | _c) & _d) | (_b & _c)) + (x) + 0x8f1bbcdc + ROL(_a, 5); _b = ROL(_b, 30);
	#define SHA1_II(_a,_b,_c,_d,_e,x,i)      _e += (_b ^ (_c ^ _d))               + (x) + 0xca62c1d6 + ROL(_a, 5); _b = ROL(_b, 30);

			SHA1_FF(a,b,c,d,e,               0, 5);
			SHA1_FF(e,a,b,c,d,               0, 6);
			SHA1_FF(d,e,a,b,c,               0, 7);
			SHA1_FF(c,d,e,a,b,               0, 8);
			SHA1_FF(b,c,d,e,a,               0, 9);
			SHA1_FF(a,b,c,d,e,               0,10);
			SHA1_FF(e,a,b,c,d,               0,11);
			SHA1_FF(d,e,a,b,c,               0,12);
			SHA1_FF(c,d,e,a,b,               0,13);
			SHA1_FF(b,c,d,e,a,               0,14);
			SHA1_FF(a,b,c,d,e,           pwLen,15);
			SHA1_FF(e,a,b,c,d,           w[ 0],16);
			SHA1_FF(d,e,a,b,c,c_r17(i) ^ w[ 1],17);
			SHA1_FF(c,d,e,a,b,           w[ 2],18);
			SHA1_FF(b,c,d,e,a,           w[ 3],19);

			SHA1_GG(a,b,c,d,e,c_r20(i) ^ w[ 4],20);
			SHA1_GG(e,a,b,c,d,           w[ 5],21);
			SHA1_GG(d,e,a,b,c,           w[ 6],22);
			SHA1_GG(c,d,e,a,b,c_r23(i) ^ w[ 7],23);
			SHA1_GG(b,c,d,e,a,           w[ 8],24);
			SHA1_GG(a,b,c,d,e,c_r20(i) ^ w[ 9],25);
			SHA1_GG(e,a,b,c,d,c_r26(i) ^ w[10],26);
			SHA1_GG(d,e,a,b,c,           w[11],27);
			SHA1_GG(c,d,e,a,b,           w[12],28);
			SHA1_GG(b,c,d,e,a,c_r29(i) ^ w[13],29);
			SHA1_GG(a,b,c,d,e,           w[14],30);
			SHA1_GG(e,a,b,c,d,c_r31(i) ^ w[15],31);
			SHA1_GG(d,e,a,b,c,c_r32(i) ^ w[16],32);
			SHA1_GG(c,d,e,a,b,c_r33(i) ^ w[17],33);
			SHA1_GG(b,c,d,e,a,           w[18],34);
			SHA1_GG(a,b,c,d,e,c_r35(i) ^ w[19],35);
			SHA1_GG(e,a,b,c,d,c_r26(i) ^ w[20],36);
			SHA1_GG(d,e,a,b,c,c_r37(i) ^ w[21],37);
			SHA1_GG(c,d,e,a,b,c_r38(i) ^ w[22],38);
			SHA1_GG(b,c,d,e,a,c_r26(i) ^ w[23],39);

			SHA1_HH(a,b,c,d,e,           w[24],40);
			SHA1_HH(e,a,b,c,d,c_r41(i) ^ w[25],41);
			SHA1_HH(d,e,a,b,c,           w[26],42);
			SHA1_HH(c,d,e,a,b,c_r43(i) ^ w[27],43);
			SHA1_HH(b,c,d,e,a,c_r44(i) ^ w[28],44);
			SHA1_HH(a,b,c,d,e,c_r45(i) ^ w[29],45);
			SHA1_HH(e,a,b,c,d,           w[30],46);
			SHA1_HH(d,e,a,b,c,c_r47(i) ^ w[31],47);
			SHA1_HH(c,d,e,a,b,c_r48(i) ^ w[32],48);
			SHA1_HH(b,c,d,e,a,c_r49(i) ^ w[33],49);
			SHA1_HH(a,b,c,d,e,c_r50(i) ^ w[34],50);
			SHA1_HH(e,a,b,c,d,c_r38(i) ^ w[35],51);
			SHA1_HH(d,e,a,b,c,c_r37(i) ^ w[36],52);
			SHA1_HH(c,d,e,a,b,c_r53(i) ^ w[37],53);
			SHA1_HH(b,c,d,e,a,           w[38],54);
			SHA1_HH(a,b,c,d,e,c_r55(i) ^ w[39],55);
			SHA1_HH(e,a,b,c,d,c_r56(i) ^ w[40],56);
			SHA1_HH(d,e,a,b,c,c_r57(i) ^ w[41],57);
			SHA1_HH(c,d,e,a,b,c_r38(i) ^ w[42],58);
			SHA1_HH(b,c,d,e,a,c_r59(i) ^ w[43],59);
			
			SHA1_II(a,b,c,d,e,c_r60(i) ^ w[44],60);
			SHA1_II(e,a,b,c,d,c_r61(i) ^ w[45],61);
			SHA1_II(d,e,a,b,c,c_r62(i) ^ w[46],62);
			SHA1_II(c,d,e,a,b,c_r63(i) ^ w[47],63);
			SHA1_II(b,c,d,e,a,c_r38(i) ^ w[48],64);
			SHA1_II(a,b,c,d,e,c_r65(i) ^ w[49],65);
			SHA1_II(e,a,b,c,d,           w[50],66);
			SHA1_II(d,e,a,b,c,c_r67(i) ^ w[51],67);
			SHA1_II(c,d,e,a,b,c_r68(i) ^ w[52],68);
			SHA1_II(b,c,d,e,a,c_r69(i) ^ w[53],69);
			SHA1_II(a,b,c,d,e,           w[54],70);

	#undef SHA1_FF
	#undef SHA1_GG
	#undef SHA1_HH
	#undef SHA1_II


			// **** 7 finish hash ****
			uint32_t tha = c_ha, thb = hb, thc = c_hc, thd = c_hd, the = c_he;

			// SHA1_II(b,c,d,e,a,    x,79);
			// hc = ROL(hc, 2);
			// ha -= (hc ^ (hd ^ he)) + 0xca62c1d6 + ROL(hb, 5);
			tha -= w[79-16] ^ c_r79(i);

			// SHA1_II(c,d,e,a,b,    x,78);
			// hd = ROL(hd, 2);
			// hb -= w[78-16] + 0xca62c1d6 + ROL(hc, 5);
			// dXORe = hd ^ he;
			thb -= c_dXORe ^ tha;

			// SHA1_II(d,e,a,b,c,    x,77);
			// heOld = he = ROL(he, 2);
			// hc -= 0xca62c1d6 + ROL(hd, 5);
			thc -= (c_heOld ^ (tha ^ thb)) + (w[77-16] ^ c_r77(i));

			// SHA1_II(e,a,b,c,d,    x,76);
			tha = ROL(tha, 2);
			// hd -= 0xca62c1d6 + ROL(he, 5);
			thd -= (tha ^ (thb ^ thc)) + (w[76-16] ^ c_r76(i));

			// SHA1_II(a,b,c,d,e,    x,75);
			thb = ROL(thb, 2);
			// he = he - 0xca62c1d6;
			the -= (thb ^ (thc ^ thd)) + (w[75-16] ^ c_r75(i)) + ROL(tha, 5);

			// ...

			// SHA1_II(d,e,a,b,c,    x,72);
			the = ROL(the, 2);


			// **** 8 check ****
			if (e == the)
			{
				if (CHEAT != 0)
				{
					d_foundPw[0] = pw0;
					d_foundPw[1] = pw1;
					d_foundPw[2] = pw2;
					return;
				}
				else
				{
					// SHA1_II(e,a,b,c,d,    x,71);
					d += (a ^ (b ^ c)) + (w[71-16] ^ c_r71(i)) + 0xca62c1d6 + ROL(e, 5);
					if (ROL(d, 30) == thd)
					{
						a = ROL(a, 30);

						// SHA1_II(d,e,a,b,c,    x,72);
						c += (e ^ (a ^ b)) + (w[72-16] ^ c_r72(i)) + 0xca62c1d6 + ROL(d, 5);
						if (ROL(c, 30) == thc)
						{
							e = ROL(e, 30);

							// SHA1_II(c,d,e,a,b,    x,73);
							b += (d ^ (e ^ a)) + (w[73-16] ^ c_r73(i)) + 0xca62c1d6 + ROL(c, 5);
							if (b == thb)
							{
								d = ROL(d, 30);

								// SHA1_II(b,c,d,e,a,    x,74);
								a += (c ^ (d ^ e)) + (w[74-16] ^ c_r74(i)) + 0xca62c1d6 + ROL(b, 5);
								if (a == tha)
								{
									d_foundPw[0] = pw0;
									d_foundPw[1] = pw1 + c_char(i) + c_char_2(j);
									if (SIZE >= 8)
									{
										d_foundPw[2] = pw2;
										if (SIZE >= 12)
										{
											d_foundPw[3] = pw3;
										}
									}
									return;
								}
							}
						}
					}
				}
			}
		}
		if (j + 1 >= char2End)
		{
			break;
		}
		td -= c_char_2(j); // c_char_2(j) = 0x00??0000
		w[17-16] ^= c_r17_2(j);
		w[20-16] ^= c_r20_2(j); // c_r20_2 = c_r25_2
		w[23-16] ^= c_r23_2(j);
		w[25-16] ^= c_r20_2(j); // c_r20_2 = c_r25_2
		w[26-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[29-16] ^= c_r29_2(j);
		w[31-16] ^= c_r31_2(j);
		w[32-16] ^= c_r32_2(j);
		w[33-16] ^= c_r33_2(j);
		w[35-16] ^= c_r35_2(j);
		w[36-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[37-16] ^= c_r37_2(j); // c_r37_2 = c_r52_2
		w[38-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[39-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[41-16] ^= c_r41_2(j);
		w[43-16] ^= c_r43_2(j);
		w[44-16] ^= c_r44_2(j);
		w[45-16] ^= c_r45_2(j);
		w[47-16] ^= c_r47_2(j);
		w[48-16] ^= c_r48_2(j);
		w[49-16] ^= c_r49_2(j);
		w[50-16] ^= c_r50_2(j);
		w[51-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[52-16] ^= c_r37_2(j); // c_r37_2 = c_r52_2
		w[53-16] ^= c_r53_2(j);
		w[55-16] ^= c_r55_2(j);
		w[56-16] ^= c_r56_2(j);
		w[57-16] ^= c_r57_2(j);
		w[58-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[59-16] ^= c_r59_2(j);
		w[60-16] ^= c_r60_2(j);
		w[61-16] ^= c_r61_2(j);
		w[62-16] ^= c_r62_2(j);
		w[63-16] ^= c_r63_2(j);
		w[64-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[65-16] ^= c_r65_2(j);
		w[67-16] ^= c_r67_2(j);
		w[68-16] ^= c_r68_2(j);
		w[69-16] ^= c_r69_2(j);
		w[71-16] ^= c_r71_2(j);
		w[72-16] ^= c_r72_2(j);
		w[73-16] ^= c_r73_2(j);
		w[74-16] ^= c_r74_2(j);
		w[75-16] ^= c_r75_2(j);
		w[76-16] ^= c_r76_2(j);
		w[77-16] ^= c_r77_2(j);
		w[79-16] ^= c_r79_2(j);
		j++;
		w[17-16] ^= c_r17_2(j);
		w[20-16] ^= c_r20_2(j); // c_r20_2 = c_r25_2
		w[23-16] ^= c_r23_2(j);
		w[25-16] ^= c_r20_2(j); // c_r20_2 = c_r25_2
		w[26-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[29-16] ^= c_r29_2(j);
		w[31-16] ^= c_r31_2(j);
		w[32-16] ^= c_r32_2(j);
		w[33-16] ^= c_r33_2(j);
		w[35-16] ^= c_r35_2(j);
		w[36-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[37-16] ^= c_r37_2(j); // c_r37_2 = c_r52_2
		w[38-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[39-16] ^= c_r26_2(j); // c_r26_2 = c_r36_2 = c_r39_2
		w[41-16] ^= c_r41_2(j);
		w[43-16] ^= c_r43_2(j);
		w[44-16] ^= c_r44_2(j);
		w[45-16] ^= c_r45_2(j);
		w[47-16] ^= c_r47_2(j);
		w[48-16] ^= c_r48_2(j);
		w[49-16] ^= c_r49_2(j);
		w[50-16] ^= c_r50_2(j);
		w[51-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[52-16] ^= c_r37_2(j); // c_r37_2 = c_r52_2
		w[53-16] ^= c_r53_2(j);
		w[55-16] ^= c_r55_2(j);
		w[56-16] ^= c_r56_2(j);
		w[57-16] ^= c_r57_2(j);
		w[58-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[59-16] ^= c_r59_2(j);
		w[60-16] ^= c_r60_2(j);
		w[61-16] ^= c_r61_2(j);
		w[62-16] ^= c_r62_2(j);
		w[63-16] ^= c_r63_2(j);
		w[64-16] ^= c_r38_2(j); // c_r38_2 = c_r51_2 = c_r58_2 = c_r64_2
		w[65-16] ^= c_r65_2(j);
		w[67-16] ^= c_r67_2(j);
		w[68-16] ^= c_r68_2(j);
		w[69-16] ^= c_r69_2(j);
		w[71-16] ^= c_r71_2(j);
		w[72-16] ^= c_r72_2(j);
		w[73-16] ^= c_r73_2(j);
		w[74-16] ^= c_r74_2(j);
		w[75-16] ^= c_r75_2(j);
		w[76-16] ^= c_r76_2(j);
		w[77-16] ^= c_r77_2(j);
		w[79-16] ^= c_r79_2(j);
	}
#undef ROL
}


// **** crackers ****
void sbSha1Kernel_cracker_any7(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
//	sbSha1_cracker_any<7><<<blocks, threads, 524>>>(d_shared, d_foundPw, pwLen, char2Start, char2End, index, maxIndex);
}

void sbSha1Kernel_cracker_any11(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
	sbSha1_cracker_any<11,0><<<blocks, threads, 524>>>(d_shared, d_foundPw, pwLen, char2Start, char2End, index, maxIndex);
}

void sbSha1Kernel_cracker_any11_cheat(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
	sbSha1_cracker_any<11,1><<<blocks, threads, 524>>>(d_shared, d_foundPw, pwLen, char2Start, char2End, index, maxIndex);
}

void sbSha1Kernel_cracker_any15(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
//	sbSha1_cracker_any<15><<<blocks, threads, 524>>>(d_shared, d_foundPw, pwLen, char2Start, char2End, index, maxIndex);
}

/*void sbSha1Kernel_cracker_any19(int blocks, int threads, uint32_t *d_shared, uint32_t *d_foundPw, uint32_t pwLen, uint32_t char2Start, uint32_t char2End, uint64_t index, uint64_t maxIndex)
{
	sbSha1_cracker_any<19><<<blocks, threads, 524>>>(d_shared, d_foundPw, pwLen, char2Start, char2End, index, maxIndex);
}*/

extern "C" int sha1_getKernel(uint32_t pwLen, kernel_cracker *cracker, kernel_cracker *crackerCheat)
{
	if (pwLen <= 4)
	{
		return 1;
	}
	else if (pwLen <= 7)
	{
		*cracker = sbSha1Kernel_cracker_any7;
	}
	else if (pwLen <= 11)
	{
		*cracker      = sbSha1Kernel_cracker_any11;
		*crackerCheat = sbSha1Kernel_cracker_any11_cheat;
	}
	else if (pwLen <= 15)
	{
		*cracker = sbSha1Kernel_cracker_any15;
	}
/*	else if (pwLen <= 19)
	{
		*cracker = sbSha1Kernel_cracker_any19;
	}*/
	else
	{
		return 1;
	}
	return 0;
}
